(******************************************************************************)
(*  ######  #####  #####  ###   ###    A Formal C Concurrency Model           *)
(*  ##     ##     ##      #### ####                                           *)
(*  #####  ##     ##      ## ### ##    Copyright (c) Viktor Vafeiadis         *)
(*  ##     ##     ##      ##     ##                                           *)
(*  ##      #####  #####  ##     ##    See LICENSE.txt for license.           *)
(******************************************************************************)

Require Import Vbase List Relations. 
Set Implicit Arguments.

Lemma clos_refl_transE A r a b :
  clos_refl_trans A r a b <-> a = b \/ clos_trans A r a b.
Proof.
  split; ins; desf; vauto; induction H; desf; vauto. 
Qed.

Lemma clos_trans_in_rt A r a b :
  clos_trans A r a b -> clos_refl_trans A r a b.
Proof.
  induction 1; vauto. 
Qed.

Lemma rt_t_trans A r a b c :
  clos_refl_trans A r a b -> clos_trans A r b c -> clos_trans A r a c.
Proof.
  ins; induction H; eauto using clos_trans. 
Qed.

Lemma t_rt_trans A r a b c :
  clos_trans A r a b -> clos_refl_trans A r b c -> clos_trans A r a c.
Proof.
  ins; induction H0; eauto using clos_trans. 
Qed.

Lemma clos_trans_mon A (r r' : relation A) a b :
  clos_trans A r a b ->
  (forall a b, r a b -> r' a b) ->
  clos_trans A r' a b.
Proof.
  induction 1; ins; eauto using clos_trans.
Qed.

Lemma clos_refl_trans_mon A (r r' : relation A) a b :
  clos_refl_trans A r a b ->
  (forall a b, r a b -> r' a b) ->
  clos_refl_trans A r' a b.
Proof.
  induction 1; ins; eauto using clos_refl_trans.
Qed.

Lemma t_step_rt A R x y : 
  clos_trans A R x y <-> exists z, R x z /\ clos_refl_trans A R z y.
Proof.
  split; ins; desf.
    by apply clos_trans_tn1 in H; induction H; desf; eauto using clos_refl_trans.
  by rewrite clos_refl_transE in *; desf; eauto using clos_trans.
Qed.

(******************************************************************************)
(** * C11 concurrency model definition *)
(******************************************************************************)

Definition immediate X (rel : relation X) (a b: X) :=
  rel a b /\ (forall c (R1: rel a c) (R2: rel c b), False).

Definition irreflexive X (rel : relation X) := forall x, rel x x -> False.

Definition acyclic X (rel : relation X) := irreflexive (clos_trans X rel).

Definition is_total X (cond: X -> Prop) (rel: relation X) :=
  forall a (IWa: cond a)
         b (IWb: cond b) (NEQ: a <> b),
    rel a b \/ rel b a.

Definition restr_subset X (cond: X -> Prop) (rel rel': relation X) :=
  forall a (IWa: cond a)
         b (IWb: cond b) (REL: rel a b),
    rel' a b.

Require Import Classical.
Require Import CoLoR.Util.Relation.Total. 

Section total_extension.

  Variable X : Set.
  Variable dom : list X.
  Variable rel : relation X.

  Definition tot_ext := LETS rel dom.

  Lemma tot_ext_irr : acyclic rel -> irreflexive tot_ext.
  Proof.
    ins; apply LETS_irrefl; repeat red; ins; try tauto.
    eapply H with x; revert H0; generalize x at 1 3; induction 1; vauto.
    red in H0; desf; vauto.
  Qed.

  Lemma tot_ext_total : is_total (fun x => In x dom) tot_ext.
  Proof.
    red; ins.
    destruct LETS_total with (R:=rel) (l:=dom) (x:=a) (y:=b); desf; eauto;
    repeat red; ins; tauto.
  Qed.

  Lemma tot_ext_trans : transitive tot_ext.
  Proof. by apply LETS_transitive. Qed.

End total_extension.